/*
 * Forced apm sleep daemon.
 * Copyright 2000 Joey Hess <joeyh@kitenet.net> under the terms of the
 * GNU GPL
 *
 * This daemon is for laptops that do not automatically go to sleep after
 * a period of inactivity, when you forgot and leave them turned on.
 */

/* Added by Andrej Cedilnik
 *  - Added support for Agenda Vr3
 *  - Changed default time to 60 sec
 *  - Added option -d for debug mode
 *  - Added support for serial port
 */

/* Modified by Chris AtLee <chris@agendacomputing.com>
 *  - Fixed problem where apm-sleep would ignore interrupts not in
 *    /proc/interrupts at the time when apm-sleep was started
 */

/* Modified by Enrique Vidal <evidal@iti.upv.es> June-2004 
 *  - added code to send an awakening event (``kill(-1, SIGCONT)'')
 *    at the point where apm-sleep awakes itself after hibernation
 */

#include <unistd.h>
#include <termios.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <signal.h>

#ifndef AGENDA_VR3
#include <linux/apm_bios.h>
#include <apm.h>
#endif /* AGENDA_VR3 */

#define MAX_IRQS 255
#define INTERRUPTS "/proc/interrupts"
#define DEFAULT_SLEEP_TIME 10
#define PID_FILE "/var/run/apm-sleep.pid"

int irqs[MAX_IRQS]; /* irqs to examine have a value of 1 */
int max_unused=120; /* in seconds - default 2 minutes */
int check_ac=0;
int ac_max_unused;

#ifdef AGENDA_VR3
char *sleep_command="echo 1 > /proc/sys/pm/sleep";
#else  /* AGENDA_VR3 */
char *sleep_command="apm -s";
#endif /* AGENDA_VR3 */

int autoprobe=1;
int daemonize=1;
int debug=0;
int have_irqs=0;
int sleep_time = DEFAULT_SLEEP_TIME;

char* watched_interrupts[] = {"mouse", "keyboard", 
#ifdef AGENDA_VR3
	"vrbuttons", "vr41xxtpanel", "serial",
#endif
	NULL};

/* Should I watch this interrupt? */
int monitor_interrupt(char* line) {
	int i;
	for (i=0; watched_interrupts[i] != NULL; ++i) {
		if (strstr(line, watched_interrupts[i]) != NULL)
			return 1;
	}
	return 0;
}

void usage () {
	fprintf(stderr, "Usage: apm-sleep [-s command] [-u n] [-U n] [-i n [-i n ..]] [-a] [-n] [-d] [-c n]\n");
}

void parse_command_line (int argc, char **argv) {
	extern char *optarg;
	struct option long_options[] = {
		{"nodaemon", 0, NULL, 'n'},
		{"unused", 1, NULL, 'u'},
		{"ac-unused", 1, NULL, 'U'},
		{"irq", 1, NULL, 'i'},
		{"help", 0, NULL, 'h'},
		{"sleep-command", 1, NULL, 's'},
		{"auto", 0, NULL, 'a'},
		{"check-period", 1, NULL, 'c'},
		{0, 0, 0, 0}
	};
	int force_autoprobe=0;
	int c=0;
	int i;

	while (c != -1) {
		c=getopt_long(argc,argv, "ndu:U:wi:hac:", long_options, NULL);
		switch (c) {
			case 's':
				sleep_command=strdup(optarg);
				break;
			case 'n':
				daemonize=0;
				break;
			case 'd':
				debug=1;
				break;
			case 'u':
				max_unused=atoi(optarg);
				break;
#ifndef AGENDA_VR3
			case 'U':
				if (apm_exists() != 0) {
					fprintf(stderr, "apm-sleep: apm support not present in kernel\n");
					exit(1);
				}
				check_ac=1;
				ac_max_unused=atoi(optarg);
				break;
#endif /* AGENDA_VR3 */
			case 'i':
				i = atoi(optarg);
				if ((i < 0) || (i >= MAX_IRQS)) {
					fprintf(stderr, "apm-sleep: bad irq number %d\n", i);
					exit(1);
				}
				irqs[atoi(optarg)]=1;
				autoprobe=0;
				have_irqs=1;
				break;
			case 'h':
				usage();
				exit(0);
				break;
			case 'a':
				force_autoprobe=1;
				break;
			case 'c':
				sleep_time=atoi(optarg);
				if (sleep_time <= 0) {
					fprintf(stderr, "apm-sleep: bad sleep time %d\n", sleep_time);
					exit(1);
				}
				break;
		}
	}
	if (optind < argc) {
		usage();
		exit(1);
	}
	if (force_autoprobe)
		autoprobe=1;
}

/* Keep checking the interrupts. As long as there is activity, do nothing. */
void main_loop (void) {
	long irq_count[MAX_IRQS]; /* holds previous counters of the irq's */
	int activity, i, sleep_now=0, total_unused=0;
	long v;
	time_t nowtime, oldtime=0;
	FILE *f;
	char line[64];
#ifndef AGENDA_VR3
	apm_info ai;
#endif /* AGENDA_VR3 */
	
	while (1) {
		activity=0;
		f=fopen(INTERRUPTS, "r");
		if (! f) {
			perror(INTERRUPTS);
			exit(1);
		}
		while(fgets(line,sizeof(line),f)) {
			if (sscanf(line,"%d: %ld", &i, &v) == 2) {
				if (autoprobe) {
					if (monitor_interrupt(line) && irq_count[i] != v)
					{
						activity = 1;
						irq_count[i] = v;
					}
				}
				else
				{
					if (irqs[i] && irq_count[i] != v)
					{
						activity = 1;
						irq_count[i] = v;
					}
				}
			}
		}
		fclose(f);
		
		if (activity) {
			total_unused = 0;
			if(debug) fprintf(stderr, "Event happened\n");		
		}
		else {
			total_unused += sleep_time;
#ifdef AGENDA_VR3
			sleep_now = total_unused >= max_unused;
#else  /* AGENDA_VR3 */
			if (check_ac && apm_read(&ai) == 0 && 
			    ai.ac_line_status == 1) {
				/* On wall power and it matters. */
				if (ac_max_unused > 0) {
					sleep_now = total_unused >= ac_max_unused;
				}
			}
			else {
				sleep_now = total_unused >= max_unused;
			}
#endif /* AGENDA_VR3 */
			if(debug && (total_unused % 10 == 0)) 
				printf("Total unused = %d\n", total_unused);
					
			if (sleep_now) {
				if (debug) {
                                   printf("Going to bed\n");
                                   fflush(stderr); tcdrain(2);
				}
				syslog(LOG_NOTICE, "forcing sleep");
				if (system(sleep_command) != 0)
				   syslog(LOG_ERR, "%s failed", sleep_command);

                                /* we are awakening here *******************/ 
				if (debug) {
                                   fprintf(stderr,"Awakening now\n");
                                   fflush(stderr); tcdrain(2);
                                }
                                /* try to awake everybody else */
                                if (kill(-1, SIGCONT) < 0)
				   perror("Sending SIGCONT");
                                /*******************************************/ 

				total_unused=0;
				oldtime=0;
				sleep_now=0;
			}
		}
		
		sleep(sleep_time);
		
		/* 
		 * Keep track of how long it's been since we were last
		 * here. If it was much longer than sleep_time, the system
		 * was probably suspended, or this program was, (or the 
		 * kernel is thrashing :-), so clear idle counter.
		 */
		nowtime=time(NULL);
		/* The 1 is a necessary fudge factor. */
		if (oldtime && nowtime - sleep_time > oldtime + 1) {
			syslog(LOG_NOTICE,
					"%i sec sleep; resetting timer",
					nowtime - oldtime);
			total_unused=0;
		}
		oldtime=nowtime;
	}
}

int get_sleeptime() {
	char label[15];
	char time[3];
	FILE* fp = fopen("/etc/apm-sleep.conf", "r");
	if(fp == 0)
		return -1; /* unable to open file */
	fscanf(fp, "%s %s", label, time);
	fclose(fp);
	if(strcmp(label, "sleeptime") == 0)
		return atoi(time);
	return -1; /* line doesn't exist */

}

void sighup_handler() {
	max_unused = get_sleeptime();
fprintf(stderr, "apm-sleeptime=%d\n", max_unused);
}

int main (int argc, char **argv) {
	FILE *f;
	int unused_time;

	signal(SIGHUP, sighup_handler);

	parse_command_line(argc, argv);
	if((unused_time = get_sleeptime()) > 0)
		max_unused = unused_time;
fprintf(stderr, "apm-sleeptime=%d\n", max_unused);
	if(unused_time == 0)
		return(0);  /* Exit if config file says apm-sleep should be off */
	
	openlog("apm-sleep", LOG_PID, LOG_DAEMON);
	
	if (! have_irqs && ! autoprobe) {
		fprintf(stderr, "No irqs specified, and none autoprobed.\n");
		exit(1);
	}

	if (daemonize) {
		if (daemon(0,0) == -1) {
			perror("daemon");
			exit(1);
		}
		if ((f=fopen(PID_FILE, "w")) == NULL) {
			syslog(LOG_WARNING, "unable to write %s", PID_FILE);
			exit(1);
		}
		else {
			fprintf(f, "%i\n", getpid());
			fclose(f);
		}
	}
	
	main_loop();
	
	return(0); // never reached
}
