// gbm.c
// Generic button manager for Linux VR
//
// Copyright (C) 2000 Michael Klar <wyldfier@iname.com>
// Copyright (C) 2000 Bradley D. LaRonde <brad@ltc.com>
//
// Modified March 2001 by Chris AtLee <chris@agendacomputing.com>
// Apr 4, 2001: BTN_SYNC handler added by Andrew Park (apark@cdf.toronto.edu)
//              It spawns a child to execute /etc/gbm/on_sync file
//
// Modified April 2001 by James Smith <james@agendacomputing.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <sys/poll.h>
#include <sys/ioctl.h>
#include <linux/fb.h>
#include <linux/buttons.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <syslog.h>

extern int detach(int nochdir, int noclose);

#define DEBUG_GBM
#define ON_SYNC "/etc/gbm/on_sync"

// should the vr3 go into hibernation?
int turn_off;
// how long to hold the power button before toggling the backlight?
int time_to_backlight = 2;
// should the unit go to sleep when the pen is inserted?
int hibernate_on_pen = 1;

int echo(const char* s, const char* filename)
{
	int fd;
	int result;

	fd = open(filename, O_WRONLY);
	if(fd < 0) {
		fprintf(stderr, "Error opening %s: %s.\n",
		filename, strerror(errno));
		return -1;
	}

	result = write(fd, s, strlen(s));
	if(result == -1) {
		fprintf(stderr, "Error writing to %s: %s.\n",
		filename, strerror(errno));
		return -1;
	}

	return 0;
}

int do_sleep()
{       int ret;
//      fprintf(stderr, "Going to bed\n"); 
//      fflush(stderr); tcdrain(2); usleep(10000);

	// Writing anything to /proc/sys/pm/sleep should cause
	// the unit to sleep (hibernate).
	ret = echo("1", "/proc/sys/pm/sleep");

	/* we are awakening here ************************************/ 
//      fprintf(stderr,"Awakening now\n"); fflush(stderr); tcdrain(2);
	if (kill(-1, SIGCONT) < 0)   /* try to awake everybody else */
	   perror("Sending SIGCONT");
	/* **********************************************************/ 
        return(ret);
}

void set_backlight(int fd, unsigned char c)
{
#ifdef DEBUG_GBM
	printf("gbm: set_backlight: %.2x\n", c);
#endif
	ioctl(fd, FBIOPUT_BACKLIGHT, c);
}

int get_backlight(int fd)
{
	int num;
	ioctl(fd, FBIOGET_BACKLIGHT, &num);
#ifdef DEBUG_GBM
	printf("gbm: get_backlight: %.2x\n", num);
#endif
	return num;
}

void alarm_handler(int data)
{
	int fd;
	switch (data) {
		case SIGALRM:
			fd = open("/dev/fb0", O_NONBLOCK);
			if (fd < 0) {
				fprintf(stderr, "Error opening /dev/fb0: %s.\n",
					strerror(errno));
				return;
			}
			// toggle the backlight
			set_backlight(fd, 1 - get_backlight(fd));
			close(fd);
			// don't turn off, we handled the button press
			turn_off = 0;
			break;
	}
}

void usage()
{
	printf("gbm, version %d.%d\n", VERSION, PATCHLEVEL);
	printf("Usage: gbm [OPTION]...\n");
	printf("Manage GPIO button events.\n");
	printf("\n");
	printf("  --backtime=TIME  seconds power button held for backlight\n");
	printf("  --pen            hibernate on pen insert\n");
	printf("  --nodetach       don't detach from the controlling tty\n");
	printf("  --help           this message\n");
	exit(0);
}

void handle_event(unsigned short buff)
{
	int pid, status;
	char *agv[4] = { "sh", "-c", ON_SYNC, 0};

	switch (buff & BTN_DATA_MASK) {
		case BTN_POWER:
			// Button driver only queues "power button down" events,
			// it does not queue "power button up" events.
			// This is really no good.
			// Some other process might have done
			// "echo 1 > /proc/sys/pm/suspend",
			// and we are waking up from that, in which case
			// we will just suspend again.
			// The solutions that I see are:
			// 1.  Eat the power-on event at the kernel driver level
			//     if it causes a wake.
			// 2.  Somehow indicate to apps that the power on button
			//     caused a wake.
			// 3.  Somehow give apps a way of checking when the last
			//     wake occurred.
			// For the Agenda VR3, we only look at BTN_POWER_GPIO,
			// avoiding the issue.
			break;

		case BTN_POWER_GPIO:
			if (buff & BTN_PRESS) {
				// Set the flag to indicate that we want
				// to go to sleep when the button is released.
				// If it turns out that we wanted to toggle the
				// backlight, then the alarm handler will set
				// this variable to 0
				turn_off = 1;
				// Set an alarm to go off in the specified 
				// number of seconds.  If the button is still 
				// pressed at that time, the state of the 
				// backlight will be toggled
				alarm(time_to_backlight);
			}
			else {
				// Disable any alarms
				alarm(0);
				// If the alarm didn't happen, then this
				// variable will still be 1, so we should
				// go to sleep
				if (turn_off == 1) {
					do_sleep();
////                                        if (kill(-1, SIGCONT) < 0) 
////                                           perror("Sending SIGCONT");
				}
			}
			break;

		case BTN_BATT_DOOR:
			// This is the battery door
		        if ((buff & BTN_STATE_MASK) == BTN_RELEASE) {
				do_sleep();
////                                if (kill(-1, SIGCONT) < 0) 
////                                    perror("Sending SIGCONT");
                        }
			break;

		case BTN_PEN:
			// This is the pen slot
		        if (hibernate_on_pen && (buff & BTN_PRESS)) {
				// The pen has been inserted
				do_sleep();
////                                if (kill(-1, SIGCONT) < 0) 
////                                   perror("Sending SIGCONT");
                        }
			break;

		case BTN_SYNC:
			if(buff & BTN_PRESS) {
				// This is the sync button being pressed
				if ((pid = fork()) < 0) {
				   fprintf(stderr, "Fork failed. Exiting...\n");
				   exit(1);
				}
				if (!pid) {
				   openlog("gbm", LOG_PID, LOG_DAEMON);
				   syslog(LOG_NOTICE,"Executing %s\n", ON_SYNC);
				   closelog();
				   setgid(500);
				   setuid(500);
				   setenv("HOME", "/home/default", 1);
				   execv("/bin/sh", agv);
				   exit(127);
				}
				openlog("gbm", LOG_PID, LOG_DAEMON);
				syslog(LOG_NOTICE, "sync button press detected,"
					" spawning a child\n");
				closelog();
			} 
			break;
#ifdef DEBUG_GBM
		default:
			fprintf(stderr,"unhandled button data = 0x%04x\n",buff);
#endif
	}
}

static void sigchld_cb(int sig) {
	int status;
	if(sig == SIGCHLD)
		wait(&status);
}

int get_penconfig() {
        char label[15]; 
        char value[2];
        FILE* fp = fopen("/etc/gbm.conf", "r");
        if(fp == 0)
                return -1; /* unable to open file */
        fscanf(fp, "%s %s", label, value);
        fclose(fp);
        if(strcmp(label, "pen_hibernate") == 0)
                return atoi(value);
        return -2;
}

static void sighup_cb() {
	int h;
        h = get_penconfig();
	if(h < 0)
		hibernate_on_pen = 1;
	else
		hibernate_on_pen = h;
}

int main(int argc, char* argv[])
{
	int fbtn;
	struct pollfd pfd;
	unsigned short buff;
	struct sigaction sig;
	int p, i, h;

	// Should gbm detach itself?
	int should_detach = 1;
	signal(SIGCHLD, sigchld_cb);
	signal(SIGHUP, sighup_cb);	

	// Get options
	h = get_penconfig();
	if(h >= 0)
		hibernate_on_pen = h;

	for (i=0; i < argc; ++i) {
		if (strstr(argv[i], "--backtime=") == argv[i]) {
			p = atoi(strstr(argv[i], "=") + 1);
			if (p > 0)
				time_to_backlight = p;
		}
		else if (strcmp(argv[i], "--pen") == 0)
			hibernate_on_pen = 1;
		else if (strcmp(argv[i], "--nodetach") == 0)
			should_detach = 0;
		else if (strcmp(argv[i], "--help") == 0)
			usage();
	}

	if(should_detach) {
		// Detach from parent process.
		if(detach(0, 0) < 0) {
			perror("detaching");
			exit(2);
		}
	}

	if ((fbtn = open("/dev/vrbuttons", O_RDONLY)) < 0) {
		fprintf(stderr, 
			"Error, result %d opening /dev/vrbuttons\n", fbtn);
		return 1;
	}

	pfd.fd = fbtn;
	pfd.events = POLLIN;

	// Set up alarm handler
	sig.sa_handler = alarm_handler;
	sig.sa_flags = 0;
	sigaction(SIGALRM, &sig, 0);

	for (;;) {
		p = poll(&pfd, 1, -1);
		if (p < 0) {
			// It's ok if poll is interrupted, it's most likely
			// the alarm
			if (errno != EINTR)
				break;
			else
				continue;
		}
		if (read(fbtn, &buff, 2) != 2)
			break;

#ifdef DEBUG_GBM
		// The following is to make sure output is printed before
		// the unit goes to sleep
		fprintf(stderr, "button event 0x%.4x\n", buff);
		fflush(stderr);
		tcdrain(2);
#endif

		handle_event(buff);
	}

#ifdef DEBUG_GBM
	printf("Exiting.\n");
#endif
	close(fbtn);
	return 1;
}
